from os import environ

MINECRAFT_POCKET_EDITION = 0
MINECRAFT_PI = 1
MINECRAFT_DESKTOP = 2

minecraftType = MINECRAFT_DESKTOP

try:
     minecraftType = int(environ['MINECRAFT_TYPE'])
except:
     pass

isPE = ( minecraftType != MINECRAFT_DESKTOP )





import socket
import select
import sys
import atexit
import os
import platform
import base64
from hashlib import md5
from util import flatten_parameters_to_string

""" @author: Aron Nieminen, Mojang AB"""

class RequestError(Exception):
    pass

class Connection:
    """Connection to a Minecraft Pi game"""
    RequestFailed = "Fail"

    def __init__(self, address=None, port=None):
        self.windows = (platform.system() == "Windows" or platform.system().startswith("CYGWIN_NT"))
        if address==None:
            try:
                 address = os.environ['MINECRAFT_API_HOST']
            except KeyError:
                 address = "localhost"
        if port==None:
            try:
                 port = int(os.environ['MINECRAFT_API_PORT'])
            except KeyError:
                 port = 4711
        if int(sys.version[0]) >= 3:
            self.send = self.send_python3
            self.send_flat = self.send_flat_python3
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.connect((address, port))
        self.readFile = self.socket.makefile("r")
        self.lastSent = ""
        if self.windows:
            atexit.register(self.close)

    def __del__(self):
        if self.windows:
            self.close()
            try:
                atexit.unregister(self.close)
            except:
                pass   

    def close(self):
        try:
            if self.windows:
                # ugly hack to block until all sending is completed
                self.sendReceive("world.getBlock",0,0,0)
        except:
            pass
        try:
            self.socket.close()
        except:
            pass
            
    @staticmethod
    def tohex(data):
        return "".join((hex(b) for b in data))
            
    def authenticate(self, username, password):
        challenge = self.sendReceive("world.getBlock",0,0,0)
        if challenge.startswith("security.challenge "):
            salt = challenge[19:].rstrip()
            auth = md5(salt+":"+username+":"+password).hexdigest()
            self.send("security.authenticate", auth)

    def drain(self):
        """Drains the socket of incoming data"""
        while True:
            readable, _, _ = select.select([self.socket], [], [], 0.0)
            if not readable:
                break
            data = self.socket.recv(1500)
            if not data:
                self.socket.close()
                raise ValueError('Socket got closed')
            e =  "Drained Data: <%s>\n"%data.strip()
            e += "Last Message: <%s>\n"%self.lastSent.strip()
            sys.stderr.write(e)
                                             
    def send(self, f, *data):
        """Sends data. Note that a trailing newline '\n' is added here"""
        s = "%s(%s)\n"%(f, flatten_parameters_to_string(data))
        #print "f,data:",f,data
        self.drain()
        self.lastSent = s
        self.socket.sendall(s)

    def send_python3(self, f, *data):
        """Sends data. Note that a trailing newline '\n' is added here"""
        s = "%s(%s)\n"%(f, flatten_parameters_to_string(data))
        #print "f,data:",f,data
        self.drain()
        self.lastSent = s
        self.socket.sendall(s.encode("utf-8"))

    def send_flat(self, f, data):
        """Sends data. Note that a trailing newline '\n' is added here"""
#        print "f,data:",f,list(data)
        s = "%s(%s)\n"%(f, ",".join(data))
        self.drain()
        self.lastSent = s
        self.socket.sendall(s)

    def send_flat_python3(self, f, data):
        """Sends data. Note that a trailing newline '\n' is added here"""
#        print "f,data:",f,list(data)
        s = "%s(%s)\n"%(f, ",".join(data))
        self.drain()
        self.lastSent = s
        self.socket.sendall(s.encode("utf-8"))

    def receive(self):
        """Receives data. Note that the trailing newline '\n' is trimmed"""
        s = self.readFile.readline().rstrip("\n")
        if s == Connection.RequestFailed:
            raise RequestError("%s failed"%self.lastSent.strip())
        return s

    def sendReceive(self, *data):
        """Sends and receive data"""
        self.send(*data)
        return self.receive()

    def sendReceive_flat(self, f, data):
        """Sends and receive data"""
        self.send_flat(f, data)
        return self.receive()



import time, random, math, os
import connection
import blocklist as bl
from util import *
from event import *
conn = connection.Connection("localhost", 4711)

LIBRARY_VERSION = 0.5

# Find the player
# players = mc.getPlayerEntityIds()

# Wait for at least one player

wait_for_player = True
while wait_for_player:
    try:
        ids = conn.sendReceive("world.getPlayerIds")
    except:
        print("Waiting for a player to connect...")
        time.sleep(1)
    else:
        wait_for_player = False

players = map(int, ids.split("|"))
player = players[0]

# BLOCKS
air = bl.AIR.id
stone = bl.STONE.id
grass = bl.GRASS.id
dirt = bl.DIRT.id
cobblestone = bl.COBBLESTONE.id
wood_planks = bl.WOOD_PLANKS.id
sapling = bl.SAPLING.id
bedrock = bl.BEDROCK.id
water_flowing = bl.WATER_FLOWING.id
water = bl.WATER.id
water_stationary = bl.WATER_STATIONARY.id
lava_flowing = bl.LAVA_FLOWING.id
lava = bl.LAVA.id
lava_stationary = bl.LAVA_STATIONARY.id
sand = bl.SAND.id
gravel = bl.GRAVEL.id
gold_ore = bl.GOLD_ORE.id
iron_ore = bl.IRON_ORE.id
coal_ore = bl.COAL_ORE.id
wood = bl.WOOD.id
leaves = bl.LEAVES.id
glass = bl.GLASS.id
lapis_lazuli_ore = bl.LAPIS_LAZULI_ORE.id
lapis_lazuli = bl.LAPIS_LAZULI_BLOCK.id
sandstone = bl.SANDSTONE.id
bed = bl.BED.id
cobweb = bl.COBWEB.id
grass_tall = bl.GRASS_TALL.id
wool = bl.WOOL.id
flower_yellow = bl.FLOWER_YELLOW.id
flower_cyan = bl.FLOWER_CYAN.id
mushroom_brown = bl.MUSHROOM_BROWN.id
mushroom_red = bl.MUSHROOM_RED.id
gold = bl.GOLD_BLOCK.id
iron = bl.IRON_BLOCK.id
stone_slab_double = bl.STONE_SLAB_DOUBLE.id
stone_slab = bl.STONE_SLAB.id
brick = bl.BRICK_BLOCK.id
tnt = bl.TNT.id
bookshelf = bl.BOOKSHELF.id
moss_stone = bl.MOSS_STONE.id
obsidian = bl.OBSIDIAN.id
torch = bl.TORCH.id
fire = bl.FIRE.id
stairs_wood = bl.STAIRS_WOOD.id
chest = bl.CHEST.id
diamond_ore = bl.DIAMOND_ORE.id
diamond = bl.DIAMOND_BLOCK.id
crafting_table = bl.CRAFTING_TABLE.id
farmland = bl.FARMLAND.id
furnace_inactive = bl.FURNACE_INACTIVE.id
furnace_active = bl.FURNACE_ACTIVE.id
door_wood = bl.DOOR_WOOD.id
ladder = bl.LADDER.id
stairs_cobblestone = bl.STAIRS_COBBLESTONE.id
door_iron = bl.DOOR_IRON.id
redstone_ore = bl.REDSTONE_ORE.id
ice = bl.ICE.id
snow = bl.SNOW_BLOCK.id
cactus = bl.CACTUS.id
clay = bl.CLAY.id
sugar_cane = bl.SUGAR_CANE.id
fence = bl.FENCE.id
glowstone = bl.GLOWSTONE_BLOCK.id
stone_brick = bl.STONE_BRICK.id
glass_pane = bl.GLASS_PANE.id
melon = bl.MELON.id
fence_gate = bl.FENCE_GATE.id
glowing_obsidian = bl.GLOWING_OBSIDIAN.id
nether_reactor_core = bl.NETHER_REACTOR_CORE.id
#monster_spawner = bl.MONSTER_SPAWNER.id
#standing_sign = bl.STANDING_SIGN_BLOCK.id
#rail = bl.RAIL.id
#lever = bl.LEVER.id
#sponge = bl.SPONGE.id
#pumpkin = bl.PUMPKIN.id
#netherrack = bl.NETHERRACK.id
#soul_sand = bl.SOUL_SAND.id
#jack = bl.JACK.id
#stained_glass = bl.STAINED_GLASS.id
#cobblestone_wall = bl.COBBLESTONE_WALL.id
#prismarine = bl.PRISMARINE.id
#sea_lantern = bl.SEA_LANTERN.id
#hay_bale = bl.HAY_BALE.id
#coal = bl.COAL_BLOCK.id
#magma = bl.MAGMA_BLOCK.id
#redstone = bl.REDSTONE_BLOCK.id
#stained_glass_pane = bl.STAINED_GLASS_PANE.id
#slime = bl.SLIME_BLOCK.id
#carpet = bl.CARPET.id
#redstone_torch = bl.REDSTONE_TORCH.id
#piston = bl.PISTON.id
#sticky_piston = bl.STICKY_PISTON.id
#dispenser = bl.DISPENSER.id
#note = bl.NOTE_BLOCK.id
#stone_pressure_plate = bl.STONE_PRESSURE_PLATE.id
#hopper = bl.HOPPER.id
#dropper = bl.DROPPER.id
#activator_rail = bl.ACTIVATOR_RAIL.id
#powered_rail = bl.POWERED_RAIL.id
#detector_rail = bl.DETECTOR_RAIL.id
#beacon = bl.BEACON.id
#emerald = bl.EMERALD_BLOCK.id
#emerald_ore = bl.EMERALD_ORE.id
#quartz = bl.QUARTZ_BLOCK.id
#barrier = bl.BARRIER.id


def getblock(name):
    return globals()[name]


def chat(text):
    conn.send("chat.post", text)


def where(target=player):
    s = conn.sendReceive("entity" + ".getTile", target)
    return Vec3(*map(int, s.split(",")))


def move(x=0, y=0, z=0, target=player, absolute=False):
    s = conn.sendReceive("entity" + ".getTile", target)
    pos = Vec3(*map(int, s.split(",")))
    if not absolute:
        x += pos.x
        y += pos.y
        z += pos.z
    conn.send("entity" + ".setTile", target, intFloor(x, y, z))


def goto(x=0, y=0, z=0, target=player):
    # s = conn.sendReceive("entity" + ".getTile", target)
    # pos = Vec3(*map(int, s.split(",")))
    # if not absolute:
    #     x += pos.x
    #     y += pos.y
    #     z += pos.z
    conn.send("entity" + ".setTile", target, intFloor(x, y, z))


def changex(x=0, target=player):
    s = conn.sendReceive("entity" + ".getTile", target)
    pos = Vec3(*map(int, s.split(",")))
    x += pos.x
    y = pos.y
    z = pos.z
    conn.send("entity" + ".setTile", target, intFloor(x, y, z))


def changey(y=0, target=player):
    s = conn.sendReceive("entity" + ".getTile", target)
    pos = Vec3(*map(int, s.split(",")))
    x = pos.x
    y += pos.y
    z = pos.z
    conn.send("entity" + ".setTile", target, intFloor(x, y, z))


def changez(z=0, target=player):
    s = conn.sendReceive("entity" + ".getTile", target)
    pos = Vec3(*map(int, s.split(",")))
    x = pos.x
    y = pos.y
    z += pos.z
    conn.send("entity" + ".setTile", target, intFloor(x, y, z))


def setx(x=0, target=player):
    s = conn.sendReceive("entity" + ".getTile", target)
    pos = Vec3(*map(int, s.split(",")))
    y = pos.y
    z = pos.z
    conn.send("entity" + ".setTile", target, intFloor(x, y, z))


def sety(y=0, target=player):
    s = conn.sendReceive("entity" + ".getTile", target)
    pos = Vec3(*map(int, s.split(",")))
    x = pos.x
    z = pos.z
    conn.send("entity" + ".setTile", target, intFloor(x, y, z))


def setz(z=0, target=player):
    s = conn.sendReceive("entity" + ".getTile", target)
    pos = Vec3(*map(int, s.split(",")))
    x = pos.x
    y = pos.y
    conn.send("entity" + ".setTile", target, intFloor(x, y, z))


def sphere(block, radius=10, x=0, y=0, z=0, absolute=False, hollow=False, target=player):
    if block is list:
        block_data = block[1]
        block = block[0]
    else:
        block_data = 0
    if not absolute:
        s = conn.sendReceive("entity" + ".getTile", target)
        pos = Vec3(*map(int, s.split(",")))
        x += pos.x
        y += pos.y
        z += pos.z
    if not hollow:
        for xd in range(radius * -1, radius):
            for yd in range(radius * -1, radius):
                for zd in range(radius * -1, radius):
                    if xd ** 2 + yd ** 2 + zd ** 2 < radius ** 2:
                        conn.send("world.setBlock", intFloor(x + xd, y + yd, z + zd, block, block_data))
    else:
        for xd in range(radius * -1, radius):
            for yd in range(radius * -1, radius):
                for zd in range(radius * -1, radius):
                    if (xd ** 2 + yd ** 2 + zd ** 2 < radius ** 2) and (xd ** 2 + yd ** 2 + zd ** 2 > (radius ** 2 - (radius * 2))):
                        conn.send("world.setBlock", intFloor(x + xd, y + yd, z + zd, block, block_data))


def circle(block,
           radius=10,
           x=0, y=0, z=0,
           direction="vertical",
           absolute=False,
           target=player):
    if block is list:
        block_data = block[1]
        block = block[0]
    else:
        block_data = 0
    if not absolute:
        s = conn.sendReceive("entity" + ".getTile", target)
        pos = Vec3(*map(int, s.split(",")))
        x += pos.x
        y += pos.y
        z += pos.z
    if direction == "vertical":
        f = 1 - radius
        ddf_x = 1
        ddf_y = -2 * radius
        xd = 0
        yd = radius
        conn.send("world.setBlock", intFloor(x, y + radius, z, block, block_data))
        conn.send("world.setBlock", intFloor(x, y - radius, z, block, block_data))
        conn.send("world.setBlock", intFloor(x + radius, y, z, block, block_data))
        conn.send("world.setBlock", intFloor(x - radius, y, z, block, block_data))
        while xd < yd:
            if f >= 0:
                yd -= 1
                ddf_y += 2
                f += ddf_y
            xd += 1
            ddf_x += 2
            f += ddf_x
            conn.send("world.setBlock", intFloor(x + xd, y + yd, z, block, block_data))
            conn.send("world.setBlock", intFloor(x - xd, y + yd, z, block, block_data))
            conn.send("world.setBlock", intFloor(x + xd, y - yd, z, block, block_data))
            conn.send("world.setBlock", intFloor(x - xd, y - yd, z, block, block_data))
            conn.send("world.setBlock", intFloor(x + yd, y + xd, z, block, block_data))
            conn.send("world.setBlock", intFloor(x - yd, y + xd, z, block, block_data))
            conn.send("world.setBlock", intFloor(x + yd, y - xd, z, block, block_data))
            conn.send("world.setBlock", intFloor(x - yd, y - xd, z, block, block_data))
    elif direction == "horizontal":
        f = 1 - radius
        ddf_x = 1
        ddf_z = -2 * radius
        xd = 0
        zd = radius
        conn.send("world.setBlock", intFloor(x, y, z + radius, block, block_data))
        conn.send("world.setBlock", intFloor(x, y, z - radius, block, block_data))
        conn.send("world.setBlock", intFloor(x + radius, y, z, block, block_data))
        conn.send("world.setBlock", intFloor(x - radius, y, z, block, block_data))
        while xd < zd:
            if f >= 0:
                zd -= 1
                ddf_z += 2
                f += ddf_z
            xd += 1
            ddf_x += 2
            f += ddf_x
            conn.send("world.setBlock", intFloor(x + xd, y, z + zd, block, block_data))
            conn.send("world.setBlock", intFloor(x - xd, y, z + zd, block, block_data))
            conn.send("world.setBlock", intFloor(x + xd, y, z - zd, block, block_data))
            conn.send("world.setBlock", intFloor(x - xd, y, z - zd, block, block_data))
            conn.send("world.setBlock", intFloor(x + zd, y, z + xd, block, block_data))
            conn.send("world.setBlock", intFloor(x - zd, y, z + xd, block, block_data))
            conn.send("world.setBlock", intFloor(x + zd, y, z - xd, block, block_data))
            conn.send("world.setBlock", intFloor(x - zd, y, z - xd, block, block_data))


def line(block, x1=0, y1=0, z1=0, x2=0, y2=0, z2=0, absolute=False, target=player):
    if block is list:
        block_data = block[1]
        block = block[0]
    else:
        block_data = 0
    if not absolute:
        s = conn.sendReceive("entity" + ".getTile", target)
        pos = Vec3(*map(int, s.split(",")))
        x1 += pos.x
        y1 += pos.y
        z1 += pos.z
        x2 = pos.x + x2
        y2 = pos.y + y2
        z2 = pos.z + z2
    # List for vertices
    vertices = []
    # If the 2 points are the same, return single vertice
    if x1 == x2 and y1 == y2 and z1 == z2:
        vertices.append(Vec3(x1, y1, z1))
    # Else get all points in edge
    else:
        dx = x2 - x1
        dy = y2 - y1
        dz = z2 - z1
        ax = abs(dx) << 1
        ay = abs(dy) << 1
        az = abs(dz) << 1
        sx = ZSGN(dx)
        sy = ZSGN(dy)
        sz = ZSGN(dz)
        x = x1
        y = y1
        z = z1
        # X dominant
        if ax >= MAX(ay, az):
            yd = ay - (ax >> 1)
            zd = az - (ax >> 1)
            loop = True
            while loop:
                vertices.append(Vec3(x, y, z))
                if x == x2:
                    loop = False
                if yd >= 0:
                    y += sy
                    yd -= ax
                if zd >= 0:
                    z += sz
                    zd -= ax
                x += sx
                yd += ay
                zd += az
        # Y dominant
        elif ay >= MAX(ax, az):
            xd = ax - (ay >> 1)
            zd = az - (ay >> 1)
            loop = True
            while loop:
                vertices.append(Vec3(x, y, z))
                if y == y2:
                    loop = False
                if xd >= 0:
                    x += sx
                    xd -= ay
                if zd >= 0:
                    z += sz
                    zd -= ay
                y += sy
                xd += ax
                zd += az
        # Z dominant
        elif az >= MAX(ax, ay):
            xd = ax - (az >> 1)
            yd = ay - (az >> 1)
            loop = True
            while loop:
                vertices.append(Vec3(x, y, z))
                if z == z2:
                    loop = False
                if xd >= 0:
                    x += sx
                    xd -= az
                if yd >= 0:
                    y += sy
                    yd -= az
                z += sz
                xd += ax
                yd += ay
    for vertex in vertices:
        conn.send("world.setBlock", intFloor(vertex.x,
                                             vertex.y,
                                             vertex.z,
                                             block,
                                             block_data))


def block(block, x=0, y=0, z=0, absolute=False, target=player):
    if block is list:
        block_data = block[1]
        block = block[0]
    else:
        block_data = 0
    if not absolute:
        s = conn.sendReceive("entity" + ".getTile", target)
        pos = Vec3(*map(int, s.split(",")))
        x += pos.x
        y += pos.y
        z += pos.z
    conn.send("world.setBlock", intFloor(x, y, z, block, block_data))


def blocks(block, x1=0, y1=0, z1=0, x=0, y=0, z=0, absolute=False, target=player):
    if block is list:
        block_data = block[1]
        block = block[0]
    else:
        block_data = 0
    if not absolute:
        s = conn.sendReceive("entity" + ".getTile", target)
        pos = Vec3(*map(int, s.split(",")))
        x1 += pos.x
        y1 += pos.y
        z1 += pos.z
        x = pos.x + x
        y = pos.y + y
        z = pos.z + z
    conn.send("world.setBlocks", intFloor(x1, y1, z1, x, y, z, block, block_data))


def cube(block, side=10, x=0, y=0, z=0, absolute=False, target=player):
    if block is list:
        block_data = block[1]
        block = block[0]
    else:
        block_data = 0
    if not absolute:
        s = conn.sendReceive("entity" + ".getTile", target)
        pos = Vec3(*map(int, s.split(",")))
        x += pos.x
        y += pos.y
        z += pos.z
    conn.send("world.setBlocks", intFloor(x, y, z, x + side - 1, y + side - 1, z + side - 1, block, block_data))


def pyramid(block, width=11, x=0, y=0, z=0, absolute=False, target=player):
    if block is list:
        block_data = block[1]
        block = block[0]
    else:
        block_data = 0
    if not absolute:
        s = conn.sendReceive("entity" + ".getTile", target)
        pos = Vec3(*map(int, s.split(",")))
        x += pos.x
        y += pos.y
        z += pos.z
    if width % 2 == 0:
        width += 1
    if width == 1:
        conn.send("world.setBlock", intFloor(x, y, z, block, block_data))
    else:
        conn.send("world.setBlocks", intFloor(x, y, z, x + width - 1, y, z + width - 1, block, block_data))
        pyramid(block, width - 2, x + 1, y + 1, z + 1, absolute=True)


def over(block, target=player):
    s = conn.sendReceive("entity" + ".getTile", target)
    pos = Vec3(*map(int, s.split(",")))
    material = int(conn.sendReceive("world.getBlock", intFloor(pos.x, pos.y - 1, pos.z)))
    if material == block:
        return True
    else:
        return False


def under(target=player):
    s = conn.sendReceive("entity" + ".getTile", target)
    pos = Vec3(*map(int, s.split(",")))
    material = int(conn.sendReceive("world.getBlock", intFloor(pos.x, pos.y - 1, pos.z)))
    return material


def what(x, y, z, absolute=False, target=player):
    if not absolute:
        s = conn.sendReceive("entity" + ".getTile", target)
        pos = Vec3(*map(int, s.split(",")))
        x += pos.x
        y += pos.y
        z += pos.z
    material = int(conn.sendReceive("world.getBlock", intFloor(x, y, z)))
    return material


def near(block, radius=10, target=player):
    s = conn.sendReceive("entity" + ".getTile", target)
    pos = Vec3(*map(int, s.split(",")))
    blocks = conn.sendReceive("world.getBlocks", intFloor(pos.x - radius,
                                                          pos.y - radius,
                                                          pos.z - radius,
                                                          pos.x + radius,
                                                          pos.y + radius,
                                                          pos.z + radius))
    blocks = map(int, blocks.split(","))
    for b in blocks:
        if b == block:
            return True
    return False


def readnumber(text=""):
    done = False
    value = 0
    while not done:
        try:
            value = int(input_from_chat(text))
            done = True
        except:
            chat("Il valore inserito non e' un numero valido")
    return value


def readstring(text=""):
    done = False
    value = 0
    while not done:
        try:
            value = input_from_chat(text)
            done = True
        except:
            chat("Il valore inserito non e' valido")
    return value


def input_from_chat(text):
    chat(text)
    read_done = False
    value = "0"
    while not read_done:
        s = conn.sendReceive("events.chat.posts")
        events = [e for e in s.split("|") if e]
        poll = [ChatEvent.Post(int(e[:e.find(",")]), e[e.find(",") + 1:]) for e in events]
        for msg in poll:
            value = msg.message
            read_done = True
            break
        time.sleep(0.10)
    return value


def polygon(block, shape=6, side=10, x=0, y=0, z=0, direction="horizontal", absolute=False, target=player):
    if direction == "horizontal":
        if block is list:
            block_data = block[1]
            block = block[0]
        else:
            block_data = 0
        if not absolute:
            s = conn.sendReceive("entity" + ".getTile", target)
            pos = Vec3(*map(int, s.split(",")))
            x = x + pos.x
            y = y + pos.y
            z = z + pos.z
        angle = 0
        i = shape
        side -= 1
        startx = x
        startz = z
        while i > 0:
            if i == 1:
                targetx = startx
                targetz = startz
            else:
                targetx = int(round(x + side * math.cos(angle), 0))
                targetz = int(round(z + side * math.sin(angle), 0))
            # Line starts here:
            # List for vertices
            vertices = []
            # If the 2 points are the same, return single vertice
            if x == targetx and y == y and z == targetz:
                vertices.append(Vec3(x, y, z))
            # Else get all points in edge
            else:
                dx = targetx - x
                dy = y - y
                dz = targetz - z
                ax = abs(dx) << 1
                ay = abs(dy) << 1
                az = abs(dz) << 1
                sx = ZSGN(dx)
                sy = ZSGN(dy)
                sz = ZSGN(dz)
                x = x
                y = y
                z = z
                # X dominant
                if ax >= MAX(ay, az):
                    yd = ay - (ax >> 1)
                    zd = az - (ax >> 1)
                    loop = True
                    while loop:
                        vertices.append(Vec3(x, y, z))
                        if x == targetx:
                            loop = False
                        if yd >= 0:
                            y += sy
                            yd -= ax
                        if zd >= 0:
                            z += sz
                            zd -= ax
                        x += sx
                        yd += ay
                        zd += az
                # Y dominant
                elif ay >= MAX(ax, az):
                    xd = ax - (ay >> 1)
                    zd = az - (ay >> 1)
                    loop = True
                    while loop:
                        vertices.append(Vec3(x, y, z))
                        if y == y:
                            loop = False
                        if xd >= 0:
                            x += sx
                            xd -= ay
                        if zd >= 0:
                            z += sz
                            zd -= ay
                        y += sy
                        xd += ax
                        zd += az
                # Z dominant
                elif az >= MAX(ax, ay):
                    xd = ax - (az >> 1)
                    yd = ay - (az >> 1)
                    loop = True
                    while loop:
                        vertices.append(Vec3(x, y, z))
                        if z == targetz:
                            loop = False
                        if xd >= 0:
                            x += sx
                            xd -= az
                        if yd >= 0:
                            y += sy
                            yd -= az
                        z += sz
                        xd += ax
                        yd += ay
            for vertex in vertices:
                conn.send("world.setBlock", intFloor(vertex.x,
                                                     vertex.y,
                                                     vertex.z,
                                                     block,
                                                     block_data))
            # line(block, x, y, z, targetx, y, targetz)
            angle += 2 * math.pi / shape
            x = targetx
            z = targetz
            i -= 1


def turtle(penblock, target=player):
    chat('Remember that class names should be Capital Letter (Turtle, not turtle)!')
    return Turtle(penblock, target)


def maze(csvpath, base=grass, wall=gold, obstacle=lava, target=player):
    # open maze csv
    f = open(csvpath, "r")
    # find player position
    s = conn.sendReceive("entity" + ".getTile", target)
    pos = Vec3(*map(int, s.split(",")))
    # define z start coordinate
    z = pos.z+1
    # for each line of the csv...
    for line in f.readlines():
        data = line.split(",")
        # restart from the original x at every loop cycle
        x = pos.x+1
        # for each cell in the list...
        for cell in data:
            if cell == "0":
                selectedblock = air
            elif cell == "2":
                selectedblock = obstacle
            else:
                selectedblock = wall
            # set the selected block
            conn.send("world.setBlock", intFloor(x, pos.y, z, selectedblock))
            conn.send("world.setBlock", intFloor(x, pos.y+1, z, selectedblock))
            # build the floor
            conn.send("world.setBlock", intFloor(x, pos.y-1, z, base))
            # move on the x axis
            x = x + 1
        # move on the z axis
        z = z + 1


#class chatListener:
    #
    #
    #def __init__(self):
        #self.start()
    #
    #def start(self) :
        #self.run = True
        #self.thread = threading.Thread(target=self.listen)
        #self.thread.start()
        #
    #def listen(self) :
        #while self.run:
            #for msg  in mc.events.pollChatPosts():
                #mc.postToChat(msg.message)
            #time.sleep(0.10)
            #
    #def exit(self) :
        #self.run = False
    #
#chatl = chatListener()


# TURTLE CLASS
class Turtle:

    SPEEDTIMES = {0: 0,
                  12: 0.001,
                  11: 0.01,
                  10: 0.1,
                  9: 0.2,
                  8: 0.3,
                  7: 0.4,
                  6: 0.5,
                  5: 0.6,
                  4: 0.7,
                  3: 0.8,
                  2: 0.9,
                  1: 1}

    def __init__(self, penblock, target=player):
        # Player
        self.player = player
        # Start position
        s = conn.sendReceive("entity" + ".getTile", target)
        self.startposition = Vec3(*map(int, s.split(",")))
        # Set turtle position
        self.position = Vec3(*map(int, s.split(",")))
        # Set turtle angles
        self.heading = 0
        self.verticalheading = 0
        # Set pen down
        self._pendown = True
        # Set pen bl to black wool
        self._penblock = bl.Block(bl.WOOL.id, 15)
        # Flying to true
        self.flying = True
        # Set speed
        self.turtlespeed = 6
        # Create turtle
        self.showturtle = True
        # Set turtle block
        self.turtleblock = bl.Block(bl.DIAMOND_BLOCK.id)
        # Draw turtle
        self.draw_turtle(int(self.position.x), int(self.position.y), int(self.position.y))
        # Previous vertical heading
        self.previous = 0
        # Last turtle
        self.last_drawn_turtle = Vec3(0, 0, 0)
        # Pen block
        self.penblock(penblock)
        # Speed
        self.speed(10)

    def forward(self, distance):
        # Get end of line
        x, y, z = self.find_point_on_sphere(self.position.x,
                                            self.position.y,
                                            self.position.z,
                                            self.heading,
                                            self.verticalheading,
                                            distance)
        # Move turtle forward
        self.move_turtle(x, y, z)

    def backward(self, distance):
        # Get end of line
        x, y, z = self.find_point_on_sphere(self.position.x,
                                            self.position.y,
                                            self.position.z,
                                            self.heading,
                                            self.verticalheading - 180,
                                            distance)
        # Move turtle forward
        self.move_turtle(x, y, z)

    def move_turtle(self, x, y, z):
        # Get blocks between current position and next
        target_x, target_y, target_z = int(x), int(y), int(z)
        # If walking, set target Y to be height of world
        if self.flying is False:
            target_y = int(conn.sendReceive(target_x, target_z))
        current_x, current_y, current_z = int(self.position.x), int(self.position.y), int(self.position.z)
        # Clear the turtle
        if self.showturtle:
            self.clear_turtle(self.last_drawn_turtle.x, self.last_drawn_turtle.y, self.last_drawn_turtle.z)
        # If speed is 0 and flying, just draw the line, else animate it
        if self.turtlespeed == 0 and self.flying:
            # Draw the line
            if self._pendown:
                line(self._penblock.id, current_x, current_y - 1, current_z, target_x, target_y - 1, target_z)
        else:
            blocks_between = getLine(current_x, current_y, current_z, target_x, target_y, target_z)
            if 215 < self.verticalheading < 315:
                self.previous = -1
                for block_between in blocks_between:
                    # If walking update the y, to be the height of the world
                    if self.flying is False:
                        block_between.y = int(conn.sendReceive(block_between.x, block_between.z))
                    # Draw the turtle
                    if self.showturtle:
                        self.draw_turtle(block_between.x, block_between.y - 2, block_between.z)
                    # Draw the pen
                    if self._pendown:
                        conn.send("world.setBlock", intFloor(block_between.x,
                                                             block_between.y - 1,
                                                             block_between.z,
                                                             self._penblock.id,
                                                             self._penblock.data))
                    # Wait
                    time.sleep(self.SPEEDTIMES[self.turtlespeed])
                    # Clear the turtle
                    if self.showturtle:
                        self.clear_turtle(block_between.x, block_between.y - 2, block_between.z)
                # Update turtle's position to be the target
                self.position.x, self.position.y, self.position.z = x, y, z
                # Draw turtle
                if self.showturtle:
                    self.draw_turtle(target_x, target_y - 2, target_z)
            elif 45 < self.verticalheading < 135:
                self.previous = 1
                for block_between in blocks_between:
                    # If walking update the y, to be the height of the world
                    if self.flying is False:
                        block_between.y = int(conn.sendReceive(block_between.x, block_between.z))
                    # Draw the turtle
                    if self.showturtle:
                        self.draw_turtle(block_between.x, block_between.y, block_between.z)
                    # Draw the pen
                    if self._pendown:
                        conn.send("world.setBlock", intFloor(block_between.x,
                                                             block_between.y - 1,
                                                             block_between.z,
                                                             self._penblock.id,
                                                             self._penblock.data))
                    # Wait
                    time.sleep(self.SPEEDTIMES[self.turtlespeed])
                    # Clear the turtle
                    if self.showturtle:
                        self.clear_turtle(block_between.x, block_between.y, block_between.z)
                # Update turtle's position to be the target
                self.position.x, self.position.y, self.position.z = x, y, z
                # Draw turtle
                if self.showturtle:
                    self.draw_turtle(target_x, target_y, target_z)
            else:
                if self.previous == -1:
                    for block_between in blocks_between:
                        # If walking update the y, to be the height of the world
                        if self.flying is False:
                            block_between.y = int(conn.sendReceive(block_between.x, block_between.z))
                        # Draw the turtle
                        if self.showturtle:
                            self.draw_turtle(block_between.x, block_between.y - 2, block_between.z)
                        if self._pendown:
                            conn.send("world.setBlock", intFloor(block_between.x,
                                                                 block_between.y - 1,
                                                                 block_between.z,
                                                                 self._penblock.id,
                                                                 self._penblock.data))
                        time.sleep(self.SPEEDTIMES[self.turtlespeed])
                        if self.showturtle:
                            self.clear_turtle(block_between.x, block_between.y - 2, block_between.z)
                    # Update turtle's position to be the target
                    self.position.x, self.position.y, self.position.z = x, y, z
                    # Draw turtle
                    if self.showturtle:
                        self.draw_turtle(target_x, target_y - 2, target_z)
                else:
                    for block_between in blocks_between:
                        # If walking update the y, to be the height of the world
                        if self.flying is False:
                            block_between.y = int(conn.sendReceive(block_between.x, block_between.z))
                        # Draw the turtle
                        if self.showturtle:
                            self.draw_turtle(block_between.x, block_between.y, block_between.z)
                        if self._pendown:
                            conn.send("world.setBlock", intFloor(block_between.x,
                                                                 block_between.y - 1,
                                                                 block_between.z,
                                                                 self._penblock.id,
                                                                 self._penblock.data))
                        time.sleep(self.SPEEDTIMES[self.turtlespeed])
                        if self.showturtle:
                            self.clear_turtle(block_between.x, block_between.y, block_between.z)
                    # Update turtle's position to be the target
                    self.position.x, self.position.y, self.position.z = x, y, z
                    # Draw turtle
                    if self.showturtle:
                        self.draw_turtle(target_x, target_y, target_z)
                self.previous = 0

    def right(self, angle):
        # Rotate turtle angle to the right
        self.heading = self.heading + angle
        if self.heading > 360:
            self.heading = self.heading - 360

    def left(self, angle):
        # Rotate turtle angle to the left
        self.heading = self.heading - angle
        if self.heading < 0:
            self.heading = self.heading + 360

    def up(self, angle):
        # Rotate turtle angle up
        self.verticalheading = self.verticalheading + angle
        if self.verticalheading > 360:
            self.verticalheading = self.verticalheading - 360
        # Turn flying on
        if self.flying is False:
            self.flying = True

    def down(self, angle):
        # Rotate turtle angle down
        self.verticalheading = self.verticalheading - angle
        if self.verticalheading < 0:
            self.verticalheading = self.verticalheading + 360
        # Turn flying on
        if self.flying is False:
            self.flying = True

    def goto(self, x=0, y=0, z=0, absolute=True):
        if not absolute:
            pos = where(player)
            # Clear the turtle
            if self.showturtle:
                self.clear_turtle(self.position.x,
                                  self.position.y,
                                  self.position.z)
            # Update the position
            self.position.x = pos.x + x
            self.position.y = pos.y + y
            self.position.z = pos.z + z
            # Draw the turtle
            if self.showturtle:
                self.draw_turtle(self.position.x,
                                 self.position.y,
                                 self.position.z)
        else:
            # Clear the turtle
            if self.showturtle:
                self.clear_turtle(self.position.x,
                                  self.position.y,
                                  self.position.z)
            # Update the position
            self.position.x = x
            self.position.y = y
            self.position.z = z
            # Draw the turtle
            if self.showturtle:
                self.draw_turtle(self.position.x,
                                 self.position.y,
                                 self.position.z)

    def setposition(self, x=0, y=0, z=0, absolute=False):
        self.goto(x, y, z, absolute)

    def move(self, x=0, y=0, z=0, absolute=False):
        self.goto(x, y, z, absolute)

    def setx(self, x):
        self.goto(x, self.position.y, self.position.z)

    def sety(self, y):
        self.goto(self.position.x, y, self.position.z)

    def setz(self, z):
        self.goto(self.position.x, self.position.y, z)

    def changex(self, x):
        self.move(x, 0, 0)

    def changey(self, y):
        self.move(0, y, 0)

    def changez(self, z):
        self.move(0, 0, z)

    def setheading(self, angle):
        self.heading = angle

    def setverticalheading(self, angle):
        self.verticalheading = angle
        # Turn flying on
        if self.flying is False:
            self.flying = True

    def gohome(self):
        self.goto(self.startposition.x,
                  self.startposition.y,
                  self.startposition.z)

    def pendown(self):
        self._pendown = True

    def penup(self):
        self._pendown = False

    def isdown(self):
        return self.pendown

    def fly(self):
        self.flying = True

    def walk(self):
        self.flying = False
        self.verticalheading = 0

    def penblock(self, blockId, blockData=0):
        self._penblock = bl.Block(blockId, blockData)

    def speed(self, turtlespeed):
        self.turtlespeed = turtlespeed

    def draw_turtle(self, x, y, z):
        # Draw turtle
        conn.send("world.setBlock", intFloor(x, y, z, self.turtleblock.id, self.turtleblock.data))
        self.last_drawn_turtle = Vec3(x, y, z)

    def clear_turtle(self, x, y, z):
        # Clear turtle
        conn.send("world.setBlock", intFloor(x, y, z, bl.AIR.id))

    def find_target_block(self, turtle_x, turtle_y, turtle_z, heading, verticalheading, distance):
        x, y, z = self.find_point_on_sphere(turtle_x, turtle_y, turtle_z, heading, verticalheading, distance)
        x = int(round(x, 0))
        y = int(round(y, 0))
        z = int(round(z, 0))
        return x, y, z

    def find_point_on_sphere(self, cx, cy, cz, horizontal_angle, vertical_angle, radius):
        x = cx + (radius * (math.cos(math.radians(vertical_angle)) * math.cos(math.radians(horizontal_angle))))
        y = cy + (radius * (math.sin(math.radians(vertical_angle))))
        z = cz + (radius * (math.cos(math.radians(vertical_angle)) * math.sin(math.radians(horizontal_angle))))
        return x, y, z

    def round_xyz(self, x, y, z):
        return int(round(x, 0)), int(round(y, 0)), int(round(z, 0))

    def round_vec3(self, position):
        return Vec3(int(position.x), int(position.y), int(position.z))



from connection import Connection,RequestError
from vec3 import Vec3
from event import BlockEvent,ChatEvent
from block import Block
import math
from os import environ
from util import flatten,floorFlatten
import security

""" Minecraft PI low level api v0.1_1

    Note: many methods have the parameter *arg. This solution makes it
    simple to allow different types, and variable number of arguments.
    The actual magic is a mix of flatten_parameters() and __iter__. Example:
    A Cube class could implement __iter__ to work in Minecraft.setBlocks(c, id).

    (Because of this, it's possible to "erase" arguments. CmdPlayer removes
     entityId, by injecting [] that flattens to nothing)

    @author: Aron Nieminen, Mojang AB"""


#def strFloor(*args):
#    return [str(int(math.floor(x))) for x in flatten(args)]

def fixPipe(s):
    return s.replace('&#124;', '|').replace('&amp;','&')

def stringToBlockWithNBT(s, pipeFix = False):
    data = s.split(",")
    id = int(data[0])
    if len(data) <= 1:
        return Block(id)
    elif len(data) <= 2:
        return Block(id,int(data[1]))
    else:
        nbt = ','.join(data[2:])
        if pipeFix:
            nbt = fixPipe(nbt)
        return Block(id,int(data[1]),nbt)

class CmdPositioner:
    """Methods for setting and getting positions"""
    def __init__(self, connection, packagePrefix):
        self.conn = connection
        self.pkg = packagePrefix

    def getBlock(self, *args):
        """Get block (x,y,z) => id:int"""
        return int(self.conn.sendReceive_flat("world.getBlock", floorFlatten(args)))

    def getPitch(self, id):
        """Get entity direction (entityId:int) => Vec3"""
        s = self.conn.sendReceive(self.pkg + ".getPitch", id)
        return float(s)

    def getRotation(self, id):
        """Get entity direction (entityId:int) => Vec3"""
        s = self.conn.sendReceive(self.pkg + ".getRotation", id)
        return float(s)

    def getDirection(self, id):
        """Get entity direction (entityId:int) => Vec3"""
        s = self.conn.sendReceive(self.pkg + ".getDirection", id)
        return Vec3(*map(float, s.split(",")))

    def getPos(self, id):
        """Get entity position (entityId:int) => Vec3"""
        s = self.conn.sendReceive(self.pkg + ".getPos", id)
        return Vec3(*map(float, s.split(",")))

    def setPos(self, id, *args):
        """Set entity position (entityId:int, x,y,z)"""
        self.conn.send(self.pkg + ".setPos", id, args)

    def setDirection(self, id, *args):
        """Set entity pitch (entityId:int, x,y,z)"""
        self.conn.send(self.pkg + ".setDirection", id, args)

    def setRotation(self, id, *args):
        """Set entity rotation (entityId:int, angle)"""
        self.conn.send(self.pkg + ".setRotation", id, args)

    def setPitch(self, id, *args):
        """Set entity pitch (entityId:int, angle)"""
        self.conn.send(self.pkg + ".setPitch", id, args)

    def getTilePos(self, id, *args):
        """Get entity tile position (entityId:int) => Vec3"""
        s = self.conn.sendReceive(self.pkg + ".getTile", id)
        return Vec3(*map(int, s.split(",")))

    def setTilePos(self, id, *args):
        """Set entity tile position (entityId:int) => Vec3"""
        self.conn.send(self.pkg + ".setTile", id, floorFlatten(*args))

    def setting(self, setting, status):
        """Set a player setting (setting, status). keys: autojump"""
        self.conn.send(self.pkg + ".setting", setting, 1 if bool(status) else 0)


class CmdEntity(CmdPositioner):
    """Methods for entities"""
    def __init__(self, connection):
        CmdPositioner.__init__(self, connection, "entity")


class CmdPlayer(CmdPositioner):
    """Methods for the host (Raspberry Pi) player"""
    def __init__(self, connection, playerId=()):
        CmdPositioner.__init__(self, connection, "player" if playerId==() else "entity")
        self.id = playerId
        self.conn = connection

    def getDirection(self):
        return CmdPositioner.getDirection(self, self.id)
    def getPitch(self):
        return CmdPositioner.getPitch(self, self.id)
    def getRotation(self):
        return CmdPositioner.getRotation(self, self.id)
    def setPitch(self, *args):
        return CmdPositioner.setPitch(self, self.id, args)
    def setRotation(self, *args):
        return CmdPositioner.setRotation(self, self.id, args)
    def setDirection(self, *args):
        return CmdPositioner.setDirection(self, self.id, args)
    def getRotation(self):
        return CmdPositioner.getRotation(self, self.id)
    def getPos(self):
        return CmdPositioner.getPos(self, self.id)
    def setPos(self, *args):
        return CmdPositioner.setPos(self, self.id, args)
    def getTilePos(self):
        return CmdPositioner.getTilePos(self, self.id)
    def setTilePos(self, *args):
        return CmdPositioner.setTilePos(self, self.id, args)

class CmdCamera:
    def __init__(self, connection):
        self.conn = connection

    def setNormal(self, *args):
        """Set camera mode to normal Minecraft view ([entityId])"""
        self.conn.send("camera.mode.setNormal", args)

    def setFixed(self):
        """Set camera mode to fixed view"""
        self.conn.send("camera.mode.setFixed")

    def setFollow(self, *args):
        """Set camera mode to follow an entity ([entityId])"""
        self.conn.send("camera.mode.setFollow", args)

    def setPos(self, *args):
        """Set camera entity position (x,y,z)"""
        self.conn.send("camera.setPos", args)


class CmdEvents:
    """Events"""
    def __init__(self, connection):
        self.conn = connection

    def clearAll(self):
        """Clear all old events"""
        self.conn.send("events.clear")

    def pollBlockHits(self):
        """Only triggered by sword => [BlockEvent]"""
        s = self.conn.sendReceive("events.block.hits")
        events = [e for e in s.split("|") if e]
        return [BlockEvent.Hit(*map(int, e.split(","))) for e in events]

    def pollChatPosts(self):
        """Triggered by posts to chat => [ChatEvent]"""
        s = self.conn.sendReceive("events.chat.posts")
        events = [fixPipe(e) for e in s.split("|") if e]
        return [ChatEvent.Post(int(e[:e.find(",")]), e[e.find(",") + 1:]) for e in events]

class Minecraft:
    """The main class to interact with a running instance of Minecraft Pi."""

    def __init__(self, connection=None, autoId=True):
        if connection:
            self.conn = connection
        else:
            self.conn = Connection()

        if security.AUTHENTICATION_USERNAME and security.AUTHENTICATION_PASSWORD:
            self.conn.authenticate(security.AUTHENTICATION_USERNAME, security.AUTHENTICATION_PASSWORD)

        self.camera = CmdCamera(self.conn)
        self.entity = CmdEntity(self.conn)
        
        self.playerId = None
        
        if autoId:
            try:
                 self.playerId = int(environ['MINECRAFT_PLAYER_ID'])
                 self.player = CmdPlayer(self.conn,playerId=self.playerId)
            except:
                try:
                    self.playerId = self.getPlayerId(environ['MINECRAFT_PLAYER_NAME'])
                    self.player = CmdPlayer(self.conn,playerId=self.playerId)
                except:
                    if security.AUTHENTICATION_USERNAME:
                        try:
                            self.playerId = self.getPlayerId(security.AUTHENTICATION_USERNAME)
                            self.player = CmdPlayer(self.conn,playerId=self.playerId)
                        except:
                            self.player = CmdPlayer(self.conn)
                    else:
                        self.player = CmdPlayer(self.conn)
        else:
            self.player = CmdPlayer(self.conn)
        
        self.events = CmdEvents(self.conn)
        self.enabledNBT = False


    def spawnEntity(self, *args):
        """Spawn entity (type,x,y,z,tags) and get its id => id:int"""
        return int(self.conn.sendReceive("world.spawnEntity", args))

    def removeEntity(self, *args):
        """Remove entity (id)"""
        self.conn.send("world.removeEntity", args)

    def getBlock(self, *args):
        """Get block (x,y,z) => id:int"""
        return int(self.conn.sendReceive_flat("world.getBlock", floorFlatten(args)))

    def getBlockWithData(self, *args):
        """Get block with data (x,y,z) => Block"""
        ans = self.conn.sendReceive_flat("world.getBlockWithData", floorFlatten(args))
        return Block(*map(int, ans.split(",")[:2]))

    def getBlockWithNBT(self, *args):
        """
        Get block with data and nbt (x,y,z) => Block (if no NBT) or (Block,nbt)
        For this to work, you first need to do setting("include_nbt_with_data",1)
        """
        if not self.enabledNBT:
            self.setting("include_nbt_with_data",1)
            self.enabledNBT = True
            try:
                ans = self.conn.sendReceive_flat("world.getBlockWithData", floorFlatten(args))
            except RequestError:
                # retry in case we had a Fail from the setting
                ans = self.conn.receive()
        else:
            ans = self.conn.sendReceive_flat("world.getBlockWithData", floorFlatten(args))
        return stringToBlockWithNBT(ans)
    """
        @TODO
    """

    def fallbackGetCuboid(self, getBlock, *args):
        (x0,y0,z0,x1,y1,z1) = map(lambda x:int(math.floor(float(x))), flatten(args))
        out = []
        for y in range(min(y0,y1),max(y0,y1)+1):
            for x in range(min(x0,x1),max(x0,x1)+1):
                for z in range(min(z0,z1),max(z0,z1)+1):
                    out.append(getBlock(x,y,z))                    
        return out
        
    def fallbackGetBlocksWithData(self, *args):
        return self.fallbackGetCuboid(self.getBlockWithData, args)

    def fallbackGetBlocks(self, *args):
        return self.fallbackGetCuboid(self.getBlock, args)

    def fallbackGetBlocksWithNBT(self, *args):
        return self.fallbackGetCuboid(self.getBlockWithNBT, args)

    def getBlocks(self, *args):
        """
        Get a cuboid of blocks (x0,y0,z0,x1,y1,z1) => [id:int]
        Packed with a y-loop, x-loop, z-loop, in this order.
        """
        try:
            ans = self.conn.sendReceive_flat("world.getBlocks", floorFlatten(args))
            return map(int, ans.split(","))
        except:
            self.getBlocks = self.fallbackGetBlocks
            return self.fallbackGetBlocks(*args)
        
    def getBlocksWithData(self, *args):
        """Get a cuboid of blocks (x0,y0,z0,x1,y1,z1) => [Block(id:int, meta:int)]"""
        try:
            ans = self.conn.sendReceive_flat("world.getBlocksWithData", floorFlatten(args))
            return [Block(*map(int, x.split(",")[:2])) for x in ans.split("|")]
        except:
            self.getBlocksWithData = self.fallbackGetBlocksWithData
            return self.fallbackGetBlocksWithData(*args)

    def getBlocksWithNBT(self, *args):
        """Get a cuboid of blocks (x0,y0,z0,x1,y1,z1) => [Block(id, meta, nbt)]"""
        try:
            if not self.enabledNBT:
                self.setting("include_nbt_with_data",1)
                self.enabledNBT = True
                try:
                    ans = self.conn.sendReceive_flat("world.getBlocksWithData", floorFlatten(args))
                except RequestError:
                    # retry in case we had a Fail from the setting
                    ans = self.conn.receive()
            else:
                ans = self.conn.sendReceive_flat("world.getBlocksWithData", floorFlatten(args))
            ans = self.conn.sendReceive_flat("world.getBlocksWithData", floorFlatten(args))
            return [stringToBlockWithNBT(x, pipeFix = True) for x in ans.split("|")]
        except:
            self.getBlocksWithNBT = self.fallbackGetBlocksWithNBT
            return self.fallbackGetBlocksWithNBT(*args)

    # must have no NBT tags in Block instance
    def setBlock(self, *args):
        """Set block (x,y,z,id,[data])"""
        self.conn.send_flat("world.setBlock", floorFlatten(args))

    def setBlockWithNBT(self, *args):
        """Set block (x,y,z,id,data,nbt)"""
        data = list(flatten(args))
        self.conn.send_flat("world.setBlock", list(floorFlatten(data[:5]))+data[5:])

    # must have no NBT tags in Block instance
    def setBlocks(self, *args):
        """Set a cuboid of blocks (x0,y0,z0,x1,y1,z1,id,[data])"""
        self.conn.send_flat("world.setBlocks", floorFlatten(args))

    def setBlocksWithNBT(self, *args):
        """Set a cuboid of blocks (x0,y0,z0,x1,y1,z1,id,data,nbt)"""
        data = list(flatten(args))
        self.conn.send_flat("world.setBlocks", list(floorFlatten(data[:8]))+data[8:])

    def getHeight(self, *args):
        """Get the height of the world (x,z) => int"""
        return int(self.conn.sendReceive_flat("world.getHeight", floorFlatten(args)))

    def getPlayerId(self, *args):
        """Get the id of the current player"""
        a = tuple(flatten(args))
        if self.playerId is not None and len(a) == 0:
            return self.playerId
        else:
            return int(self.conn.sendReceive_flat("world.getPlayerId", flatten(args)))

    def getPlayerEntityIds(self):
        """Get the entity ids of the connected players => [id:int]"""
        ids = self.conn.sendReceive("world.getPlayerIds")
        return map(int, ids.split("|"))

    def saveCheckpoint(self):
        """Save a checkpoint that can be used for restoring the world"""
        self.conn.send("world.checkpoint.save")

    def restoreCheckpoint(self):
        """Restore the world state to the checkpoint"""
        self.conn.send("world.checkpoint.restore")

    def postToChat(self, msg):
        """Post a message to the game chat"""
        self.conn.send("chat.post", msg)

    def setting(self, setting, status):
        """Set a world setting (setting, status). keys: world_immutable, nametags_visible"""
        self.conn.send("world.setting", setting, 1 if bool(status) else 0)

    @staticmethod
    def create(address = None, port = None):
        return Minecraft(Connection(address, port))


if __name__ == "__main__":
    mc = Minecraft.create()
    mc.postToChat("Hello, Minecraft!")



ITEM = "Item"
XPORB = "XPOrb"
LEASHKNOT = "LeashKnot"
PAINTING = "Painting"
ARROW = "Arrow"
SNOWBALL = "Snowball"
FIREBALL = "Fireball"
SMALLFIREBALL = "SmallFireball"
THROWNENDERPEARL = "ThrownEnderpearl"
EYEOFENDERSIGNAL = "EyeOfEnderSignal"
THROWNPOTION = "ThrownPotion"
THROWNEXPBOTTLE = "ThrownExpBottle"
ITEMFRAME = "ItemFrame"
WITHERSKULL = "WitherSkull"
PRIMEDTNT = "PrimedTnt"
FALLINGSAND = "FallingSand"
FIREWORKSROCKETENTITY = "FireworksRocketEntity"
ARMORSTAND = "ArmorStand"
BOAT = "Boat"
MINECARTRIDEABLE = "MinecartRideable"
MINECARTCHEST = "MinecartChest"
MINECARTFURNACE = "MinecartFurnace"
MINECARTTNT = "MinecartTNT"
MINECARTHOPPER = "MinecartHopper"
MINECARTSPAWNER = "MinecartSpawner"
MINECARTCOMMANDBLOCK = "MinecartCommandBlock"
MOB = "Mob"
MONSTER = "Monster"
CREEPER = "Creeper"
SKELETON = "Skeleton"
SPIDER = "Spider"
GIANT = "Giant"
ZOMBIE = "Zombie"
SLIME = "Slime"
GHAST = "Ghast"
PIGZOMBIE = "PigZombie"
ENDERMAN = "Enderman"
CAVESPIDER = "CaveSpider"
SILVERFISH = "Silverfish"
BLAZE = "Blaze"
LAVASLIME = "LavaSlime"
ENDERDRAGON = "EnderDragon"
WITHERBOSS = "WitherBoss"
BAT = "Bat"
WITCH = "Witch"
ENDERMITE = "Endermite"
GUARDIAN = "Guardian"
PIG = "Pig"
SHEEP = "Sheep"
COW = "Cow"
CHICKEN = "Chicken"
SQUID = "Squid"
WOLF = "Wolf"
MUSHROOMCOW = "MushroomCow"
SNOWMAN = "SnowMan"
OZELOT = "Ozelot"
VILLAGERGOLEM = "VillagerGolem"
HORSE = "EntityHorse"
RABBIT = "Rabbit"
VILLAGER = "Villager"
ENDERCRYSTAL = "EnderCrystal"
PLAYER = "(ThePlayer)"



from distutils.core import setup
setup(
    name='pycraft_minetest',
    packages=['pycraft_minetest'],
    version='0.5',
    description='Modified, simplified and improved libraries to code in Python via Minetest.',
    author='alenorfo & gmenegoz',
    author_email='ale.norfo@gmail.com',
    url='https://github.com/sprintingkiwi/pycraft_lib',
    download_url='https://github.com/sprintingkiwi/pycraft_lib/tarball/0.5',
    keywords=['game', 'development', 'learning', 'education', 'turtle'],
    classifiers=[],
)



import settings

class Block:
    """Minecraft PI block description. Can be sent to Minecraft.setBlock/s"""
    def __init__(self, id, data=0, nbt=None):
        self.id = id
        self.data = data
        if nbt is not None and len(nbt)==0:
            self.nbt = None
        else:
            self.nbt = nbt

    def __eq__(self, rhs):
        try:
            return self.id == rhs.id and self.data == rhs.data and self.nbt == rhs.nbt
        except:
            return self.data == 0 and self.nbt is None and self.id == rhs

    def __ne__(self, rhs):
        return not self.__eq__(rhs)

    def __hash__(self):
        h = (self.id << 8) + self.data
        if self.nbt is not None:
            h ^= hash(self.nbt)

    def withData(self, data):
        return Block(self.id, data)

    def __iter__(self):
        """Allows a Block to be sent whenever id [and data] is needed"""
        if self.nbt is not None:
           return iter((self.id, self.data, self.nbt))
        else:
           return iter((self.id, self.data))

    def __repr__(self):
        if self.nbt is None:
            return "Block(%d, %d)"%(self.id, self.data)
        else:
            return "Block(%d, %d, %s)"%(self.id, self.data, repr(self.nbt))



AIR                 = Block(0)
STONE               = Block(1)
GRASS               = Block(2)
DIRT                = Block(3)
COBBLESTONE         = Block(4)
WOOD_PLANKS         = Block(5)
SAPLING             = Block(6)
BEDROCK             = Block(7)
WATER_FLOWING       = Block(8)
WATER               = WATER_FLOWING
WATER_STATIONARY    = Block(9)
LAVA_FLOWING        = Block(10)
LAVA                = LAVA_FLOWING
LAVA_STATIONARY     = Block(11)
SAND                = Block(12)
GRAVEL              = Block(13)
GOLD_ORE            = Block(14)
IRON_ORE            = Block(15)
COAL_ORE            = Block(16)
WOOD                = Block(17)
LEAVES              = Block(18)
GLASS               = Block(20)
LAPIS_LAZULI_ORE    = Block(21)
LAPIS_LAZULI_BLOCK  = Block(22)
SANDSTONE           = Block(24)
BED                 = Block(26)
COBWEB              = Block(30)
GRASS_TALL          = Block(31)
WOOL                = Block(35)
FLOWER_YELLOW       = Block(37)
FLOWER_CYAN         = Block(38)
MUSHROOM_BROWN      = Block(39)
MUSHROOM_RED        = Block(40)
GOLD_BLOCK          = Block(41)
IRON_BLOCK          = Block(42)
STONE_SLAB_DOUBLE   = Block(43)
STONE_SLAB          = Block(44)
BRICK_BLOCK         = Block(45)
TNT                 = Block(46)
BOOKSHELF           = Block(47)
MOSS_STONE          = Block(48)
OBSIDIAN            = Block(49)
TORCH               = Block(50)
FIRE                = Block(51)
STAIRS_WOOD         = Block(53)
CHEST               = Block(54)
DIAMOND_ORE         = Block(56)
DIAMOND_BLOCK       = Block(57)
CRAFTING_TABLE      = Block(58)
FARMLAND            = Block(60)
FURNACE_INACTIVE    = Block(61)
FURNACE_ACTIVE      = Block(62)
DOOR_WOOD           = Block(64)
LADDER              = Block(65)
STAIRS_COBBLESTONE  = Block(67)
DOOR_IRON           = Block(71)
REDSTONE_ORE        = Block(73)
STONE_BUTTON        = Block(77)
SNOW                = Block(78)
ICE                 = Block(79)
SNOW_BLOCK          = Block(80)
CACTUS              = Block(81)
CLAY                = Block(82)
SUGAR_CANE          = Block(83)
FENCE               = Block(85)
GLOWSTONE_BLOCK     = Block(89)
BEDROCK_INVISIBLE   = Block(95)
if settings.isPE:
   STAINED_GLASS = WOOL
else:
   STAINED_GLASS = Block(95)
STONE_BRICK         = Block(98)
GLASS_PANE          = Block(102)
MELON               = Block(103)
FENCE_GATE          = Block(107)
WOOD_BUTTON         = Block(143)
REDSTONE_BLOCK      = Block(152)
QUARTZ_BLOCK        = Block(155)

if settings.isPE:
   HARDENED_CLAY_STAINED = WOOL
else:
   HARDENED_CLAY_STAINED = Block(159)

if settings.isPE:
   SEA_LANTERN         = Block(246) # glowing obsidian
else:
   SEA_LANTERN         = Block(169)

CARPET              = Block(171)
COAL_BLOCK          = Block(173)

if settings.isPE:
   GLOWING_OBSIDIAN    = Block(246)
   NETHER_REACTOR_CORE = Block(247)
   REDSTONE_LAMP_INACTIVE = OBSIDIAN
   REDSTONE_LAMP_ACTIVE = GLOWING_OBSIDIAN
else:
   GLOWING_OBSIDIAN    = SEA_LANTERN
   NETHER_REACTOR_CORE = SEA_LANTERN
   REDSTONE_LAMP_INACTIVE = Block(123)
   REDSTONE_LAMP_ACTIVE   = Block(124)

SUNFLOWER  = Block(175,0)
LILAC      = Block(175,1)
DOUBLE_TALLGRASS = Block(175,2)
LARGE_FERN       = Block(175,3)
ROSE_BUSH        = Block(175,4)
PEONY            = Block(175,5)

WOOL_WHITE = Block(WOOL.id, 0)
WOOL_ORANGE = Block(WOOL.id, 1)
WOOL_MAGENTA = Block(WOOL.id, 2)
WOOL_LIGHT_BLUE = Block(WOOL.id, 3)
WOOL_YELLOW = Block(WOOL.id, 4)
WOOL_LIME = Block(WOOL.id, 5)
WOOL_PINK = Block(WOOL.id, 6)
WOOL_GRAY = Block(WOOL.id, 7)
WOOL_LIGHT_GRAY = Block(WOOL.id, 8)
WOOL_CYAN = Block(WOOL.id, 9)
WOOL_PURPLE = Block(WOOL.id, 10)
WOOL_BLUE = Block(WOOL.id, 11)
WOOL_BROWN = Block(WOOL.id, 12)
WOOL_GREEN = Block(WOOL.id, 13)
WOOL_RED = Block(WOOL.id, 14)
WOOL_BLACK = Block(WOOL.id, 15)

CARPET_WHITE = Block(CARPET.id, 0)
CARPET_ORANGE = Block(CARPET.id, 1)
CARPET_MAGENTA = Block(CARPET.id, 2)
CARPET_LIGHT_BLUE = Block(CARPET.id, 3)
CARPET_YELLOW = Block(CARPET.id, 4)
CARPET_LIME = Block(CARPET.id, 5)
CARPET_PINK = Block(CARPET.id, 6)
CARPET_GRAY = Block(CARPET.id, 7)
CARPET_LIGHT_GRAY = Block(CARPET.id, 8)
CARPET_CYAN = Block(CARPET.id, 9)
CARPET_PURPLE = Block(CARPET.id, 10)
CARPET_BLUE = Block(CARPET.id, 11)
CARPET_BROWN = Block(CARPET.id, 12)
CARPET_GREEN = Block(CARPET.id, 13)
CARPET_RED = Block(CARPET.id, 14)
CARPET_BLACK = Block(CARPET.id, 15)

STAINED_GLASS_WHITE = Block(STAINED_GLASS.id, 0)
STAINED_GLASS_ORANGE = Block(STAINED_GLASS.id, 1)
STAINED_GLASS_MAGENTA = Block(STAINED_GLASS.id, 2)
STAINED_GLASS_LIGHT_BLUE = Block(STAINED_GLASS.id, 3)
STAINED_GLASS_YELLOW = Block(STAINED_GLASS.id, 4)
STAINED_GLASS_LIME = Block(STAINED_GLASS.id, 5)
STAINED_GLASS_PINK = Block(STAINED_GLASS.id, 6)
STAINED_GLASS_GRAY = Block(STAINED_GLASS.id, 7)
STAINED_GLASS_LIGHT_GRAY = Block(STAINED_GLASS.id, 8)
STAINED_GLASS_CYAN = Block(STAINED_GLASS.id, 9)
STAINED_GLASS_PURPLE = Block(STAINED_GLASS.id, 10)
STAINED_GLASS_BLUE = Block(STAINED_GLASS.id, 11)
STAINED_GLASS_BROWN = Block(STAINED_GLASS.id, 12)
STAINED_GLASS_GREEN = Block(STAINED_GLASS.id, 13)
STAINED_GLASS_RED = Block(STAINED_GLASS.id, 14)
STAINED_GLASS_BLACK = Block(STAINED_GLASS.id, 15)

HARDENED_CLAY_STAINED_WHITE = Block(HARDENED_CLAY_STAINED.id, 0)
HARDENED_CLAY_STAINED_ORANGE = Block(HARDENED_CLAY_STAINED.id, 1)
HARDENED_CLAY_STAINED_MAGENTA = Block(HARDENED_CLAY_STAINED.id, 2)
HARDENED_CLAY_STAINED_LIGHT_BLUE = Block(HARDENED_CLAY_STAINED.id, 3)
HARDENED_CLAY_STAINED_YELLOW = Block(HARDENED_CLAY_STAINED.id, 4)
HARDENED_CLAY_STAINED_LIME = Block(HARDENED_CLAY_STAINED.id, 5)
HARDENED_CLAY_STAINED_PINK = Block(HARDENED_CLAY_STAINED.id, 6)
HARDENED_CLAY_STAINED_GRAY = Block(HARDENED_CLAY_STAINED.id, 7)
HARDENED_CLAY_STAINED_LIGHT_GRAY = Block(HARDENED_CLAY_STAINED.id, 8)
HARDENED_CLAY_STAINED_CYAN = Block(HARDENED_CLAY_STAINED.id, 9)
HARDENED_CLAY_STAINED_PURPLE = Block(HARDENED_CLAY_STAINED.id, 10)
HARDENED_CLAY_STAINED_BLUE = Block(HARDENED_CLAY_STAINED.id, 11)
HARDENED_CLAY_STAINED_BROWN = Block(HARDENED_CLAY_STAINED.id, 12)
HARDENED_CLAY_STAINED_GREEN = Block(HARDENED_CLAY_STAINED.id, 13)
HARDENED_CLAY_STAINED_RED = Block(HARDENED_CLAY_STAINED.id, 14)
HARDENED_CLAY_STAINED_BLACK = Block(HARDENED_CLAY_STAINED.id, 15)

LEAVES_OAK_DECAYABLE = Block(LEAVES.id, 0)
LEAVES_SPRUCE_DECAYABLE = Block(LEAVES.id, 1)
LEAVES_BIRCH_DECAYABLE = Block(LEAVES.id, 2)
LEAVES_JUNGLE_DECAYABLE = Block(LEAVES.id, 3)
LEAVES_OAK_PERMANENT = Block(LEAVES.id, 4)
LEAVES_SPRUCE_PERMANENT = Block(LEAVES.id, 5)
LEAVES_BIRCH_PERMANENT = Block(LEAVES.id, 6)
LEAVES_JUNGLE_PERMANENT = Block(LEAVES.id, 7)
if settings.isPE:
    LEAVES_ACACIA_DECAYABLE = Block(161,0)
    LEAVES_DARK_OAK_DECAYABLE = Block(161,1)
    LEAVES_ACACIA_PERMANENT = Block(161,2)
    LEAVES_DARK_OAK_PERMANENT = Block(161,3)
else:
    LEAVES_ACACIA_DECAYABLE = LEAVES_OAK_DECAYABLE
    LEAVES_DARK_OAK_DECAYABLE = LEAVES_JUNGLE_DECAYABLE
    LEAVES_ACACIA_PERMANENT = LEAVES_OAK_PERMANENT
    LEAVES_DARK_OAK_PERMANENT = LEAVES_JUNGLE_PERMANENT



import collections
import math
import copy
from vec3 import Vec3

def intFloor(*args):
    return [int(math.floor(x)) for x in flatten(args)]

def flatten(l):
    for e in l:
        if isinstance(e, collections.Iterable) and not isinstance(e, basestring):
            for ee in flatten(e): yield ee
        else: yield e


# this is highly optimized to iterables consisting at base level of ints and floats only
def floorFlatten(l):
    for e in l:
        if isinstance(e, int):
            yield str(e)
        elif isinstance(e, float):
            yield str(int(math.floor(e)))
        elif not e is None:
            for ee in floorFlatten(e): yield ee


def flatten_parameters_to_string(l):
    return ",".join(map(str, flatten(l)))


# return maximum of 2 values
def MAX(a, b):
    if a > b:
        return a
    else:
        return b


# return step
def ZSGN(a):
    if a < 0:
        return -1
    elif a > 0:
        return 1
    elif a == 0:
        return 0


def getLine(x1, y1, z1, x2, y2, z2):

        # list for vertices
        vertices = []

        # if the 2 points are the same, return single vertice
        if (x1 == x2 and y1 == y2 and z1 == z2):
            vertices.append(Vec3(x1, y1, z1))

        # else get all points in edge
        else:

            dx = x2 - x1
            dy = y2 - y1
            dz = z2 - z1

            ax = abs(dx) << 1
            ay = abs(dy) << 1
            az = abs(dz) << 1

            sx = ZSGN(dx)
            sy = ZSGN(dy)
            sz = ZSGN(dz)

            x = x1
            y = y1
            z = z1

            # x dominant
            if (ax >= MAX(ay, az)):
                yd = ay - (ax >> 1)
                zd = az - (ax >> 1)
                loop = True
                while(loop):
                    vertices.append(Vec3(x, y, z))
                    if (x == x2):
                        loop = False
                    if (yd >= 0):
                        y += sy
                        yd -= ax
                    if (zd >= 0):
                        z += sz
                        zd -= ax
                    x += sx
                    yd += ay
                    zd += az
            # y dominant
            elif (ay >= MAX(ax, az)):
                xd = ax - (ay >> 1)
                zd = az - (ay >> 1)
                loop = True
                while(loop):
                    vertices.append(Vec3(x, y, z))
                    if (y == y2):
                        loop=False
                    if (xd >= 0):
                        x += sx
                        xd -= ay
                    if (zd >= 0):
                        z += sz
                        zd -= ay
                    y += sy
                    xd += ax
                    zd += az
            # z dominant
            elif(az >= MAX(ax, ay)):
                xd = ax - (az >> 1)
                yd = ay - (az >> 1)
                loop = True
                while(loop):
                    vertices.append(Vec3(x, y, z))
                    if (z == z2):
                        loop=False
                    if (xd >= 0):
                        x += sx
                        xd -= az
                    if (yd >= 0):
                        y += sy
                        yd -= az
                    z += sz
                    xd += ax
                    yd += ay

        return vertices

#def drawPoint3d(self, x, y, z, blockType, blockData=0):
    #self.conn.send("world.setBlock", intFloor(x, y, z, blockType, blockData))

#def getHeight(self, *args):
    #"""Get the height of the world (x,z) => int"""
    #return int(self.conn.sendReceive("world.getHeight", intFloor(args)))


# def drawVertices(self, vertices, blockType, blockData=0):
#     for vertex in vertices:
#         conn.send("world.setBlock", intFloor(vertex.x,
#                                              vertex.y,
#                                              vertex.z,
#                                              blockType,
#                                              blockData))
